arXiv is now an independent nonprofit! Learn more
License: CC BY 4.0
arXiv:2302.09005v1 [cs.MS] 17 Feb 2023

GPU Offloading in ExaHyPE Through C++ Standard Algorithms

Uzmar Gomez Affiliation: Department of Computer Science, Durham University, United Kingdom Email: [    Gonzalo Brito Gadeschi Affiliation: NVIDIA, Munich, Germany    Tobias Weinzierl Affiliation: Department of Computer Science, Durham University, United Kingdom
Abstract

The ISO C++17 standard introduces parallel algorithms, a parallel programming model promising portability across a wide variety of parallel hardware including multi-core CPUs, GPUs, and FPGAs. Since 2019, the NVIDIA HPC SDK compiler suite supports this programming model for multi-core CPUs and GPUs. ExaHyPE is a solver engine for hyperbolic partial differential equations for complex wave phenomena. It supports multiple numerical methods including Finite Volumes and ADER-DG, and employs adaptive mesh refinement with dynamic load balancing via space-filling curves as well as task-based parallelism and offloading to GPUs. This study ports ExaHyPE’s tasks over blocks of Finite Volumes to the ISO C++ parallel algorithms programming model, and compares its performance and usability against an OpenMP implementation with offloading via OpenMP target directives. It shows that ISO C++ is a feasible programming model for non-trivial applications like our task-based AMR code. The realisation is bare of vendor-specific or non-C++ extensions. It however is slower than its OpenMP counterpart.

Uzmar Gomez]uzmar.gomez@hotmail.com

1 Introduction

Graphics Processing Units (GPUs) have become the main accelerator for data-parallel processing tasks driving Exascale systems [1]. Yet, porting applications to GPUs and designing GPU algorithms remains challenging. The main HPC programming paradigms for GPU-accelerated systems are ISO standard languages (C++ and Fortran) that run on GPUs “automatically”, annotation-based programming models like OpenMP and OpenACC, and GPU programming languages or extensions like CUDA or SYCL. The different approaches have pros and cons.

The NVIDIA HPC SDK supports automatic offloading of ISO C++ parallel algorithms to multi-core CPUs and GPUs [2]. Recent publications demonstrate good performance for mini-applications [3] and NVIDIA’s documentation enlists success stories for larger applications, such as the multi-physics applications m-AIA [4] and STLBM [5].

The goal of this study is to examine the feasibility of parallelizing a large application to portably target multi-core CPU and multi-GPU systems using plain ISO C++. ExaHyPE is a solver engine for hyperpolic PDE equations [6] for complex wave phenomena. It is a large and complex application with support for adaptive mesh refinement with dynamic load balancing via space-filling curves. On top of this MPI+X parallelism for the CPU, it can phrase the updates of individual cells as tasks: We restrict ourselves to the computational kernels of ExaHyPE’s Finite Volume solver for small Cartesian patches in a block-structured AMR context [7]—these are embedded into the cells and hance form tasks—for which pre-existing OpenMP offloading and SyCL versions are available. More complex offloading strategies—notably porting execution logic to a GPU—are beyond scope.

The main contributions of this work are discovering how to map individual tasks to the new ISO C++ parallel algorithms programming model, and comparing the productivity, usability, and performance of this programming model against OpenMP. We also report on the technical challenges we encountered using this novel programming model for offloading ISO C++ to GPUs.

This study is organized as follows. Section 2 describes the problem domain, the numerical scheme, and the ExaHyPE software. Section 3 discusses the core compute kernel C++ implementation, GPU execution policies, and implementation challenges. Section 4 shows our preliminary results from which we draw our conclusions.

2 Block-structured Finite Volumes in ExaHyPE

ExaHyPE is an engine to write simulation codes for hyperbolic, first-order equation systems over functions q→​(x→,t)=(q1,…,qs)\vec{q}\left(\vec{x},t\right)=\left(q_{1},\dots,q_{s}\right)

∂q→∂t+∑j=1d∂∂xj​fj→​(q→)=0,\frac{\partial\vec{q}}{\partial t}+\sum_{j=1}^{d}\frac{\partial}{\partial x_{j}}\vec{f^{j}}(\vec{q})=0, (1)

where x→∈ℝd\vec{x}\in\mathbb{R}^{d} and the directional terms fj→\vec{f^{j}} form the flux function. In ExaHyPE, users adopt such generic PDE formulations by injecting function implementations evaluating fj​(q→)f^{j}(\vec{q}). How these function evaluations are used within a numerical scheme, on which core, and when, is hidden from the users. This is the Hollywood principle from [8]. For the present experiments, we benchmark the Euler equations. In conservative form, these equations read as

∂∂t​(ρj→Et)+∇⋅(j→1ρ​j→⊗j→+p​I(Et+p)​1ρ)=0→,\frac{\partial}{\partial t}\begin{pmatrix}\rho\\ \vec{j}\\ E_{t}\end{pmatrix}+\nabla\cdot\begin{pmatrix}\vec{j}\\ \frac{1}{\rho}\vec{j}\otimes\vec{j}+pI\\ \left(E_{t}+p\right)\frac{1}{\rho}\end{pmatrix}=\vec{0}, (2)

where j→=ρ​u→\vec{j}=\rho\vec{u} is the momentum density, ρ\rho is the fluid mass density, pp is the pressure, and EtE_{t} is the total energy density.

ExaHyPE’s spatial discretisation is fixed: The code employs a block-structured mesh [9], resulting from spacetrees [8]: A cubic computational domain is split into three chunks along each coordinate axis and this process then is repeated locally and recursively. We end up with a dynamically adaptive grid of cubes. On top of this spatial discretisation, ExaHyPE offers various numerical schemes. We stick to block-structured Finite Volumes here: Each cube of the grid hosts a small Cartesian patch of p×pp\times p or p×p×pp\times p\times p volumes. Each patch is supplemented with a halo layer.

It is ExaHyPE’s responsibility to maintain, store and traverse the block-structured mesh. Per Cartesian patch, the code invokes an update kernel to advance a patch in time. This kernel is tailored towards the PDE of interest by the user-provided flux functions. After that, it is ExaHyPE’s responsibility to keep all (halo) data consistent.

Parallelisation and GPU offloading

ExaHyPE employs a traditional non-overlapping domain decomposition realised through MPI. Per MPI rank, it then applies a non-overlapping domain decomposition again. The resulting subdomains are distributed per rank among the threads.

Each thread runs through its subdomain per time step and advances the numerical solution. Throughout the traversal, the code prioritises patches along AMR boundaries or the subdomain boundaries: These patches are updated immediately while we run throught the domain, as their output is urgent as it feeds into domain boundary exchange or is involved in potentially expensive AMR routines.

The remaining patches are held back and put into a thread-local buffer. Whenever this buffer exceeds a given threshold NN, we take the NN patches and deploy them as a whole to an accelerator. All remaining patches are deployed to normal tasks at the end of a traversal to exploit the host cores. The classification of patches realises the principle of enclave tasking [10], while the buffering uses a concept from [7].

A generic Finite Volume kernel

Per volume 𝒞i\mathcal{C}_{i} within each p×p×pp\times p\times p patch, we use a Finite Difference approximation of the time derivative in (1) subject to a Finite Volume scheme for the spatial discretisation:

Qin+1−QinΔ​t+Fi+1/2n−Fi−1/2nΔ​x=0,\frac{Q_{i}^{n+1}-Q_{i}^{n}}{\Delta t}+\frac{F_{i+1/2}^{n}-F_{i-1/2}^{n}}{\Delta x}=0, (3)

where Fi−1/2nF_{i-1/2}^{n} is the flux over the faces of 𝒞i\mathcal{C}_{i}. Subsequently, we use Rusanov’s approximation to the flux from (3):

Fi−1/2n=12​[f⁡(Qi−1n)+f⁡(Qin)−ai−1/2​(Qin−Qi−1n)]F_{i-1/2}^{n}=\frac{1}{2}\left[f\left(Q_{i-1}^{n}\right)+f\left(Q_{i}^{n}\right)-a_{i-1/2}\left(Q_{i}^{n}-Q_{i-1}^{n}\right)\right] (4)

where

ai−1/2\displaystyle a_{i-1/2} =\displaystyle= max(|f′(q)|)over all q between Qi−1n and Qin\displaystyle\max\left(\left|f^{\prime}(q)\right|\right)\ \text{over all }q\text{ between }Q_{i-1}^{n}\text{ and }Q_{i}^{n} (5)
=\displaystyle= max⁡(|λk|).\displaystyle\max\left(\left|\lambda_{k}\right|\right).

λk\lambda_{k} are the eigenvalues of the PDE’s Jacobian matrix.

With this generic numerical scheme, we can implement a generic patch update kernel which is tailored towards (2) by setting the number of quantities of interest to d+2d+2 and by injecting two routines: one evaluates Euler’s flux, the other one returns the maximum eigenvalue.

3 A C++ GPU computer kernel

Algorithm 1 Compute steps for one patch. A kernel for multiple patches in one rush adds an additional outer loop.
1: Qnew=QoldQ^{\text{new}}=Q^{\text{old}}
2: Compute ai−1/2​Qolda_{i-1/2}Q^{\text{old}} where ai−1/2a_{i-1/2} is given by eq. 5.
3: a=max⁡(λmax​Qold)a=\max\left(\lambda_{\max}Q^{\text{old}}\right) for face in boundary of QoldQ^{\text{old}}.
4: Update the volume by a⁡(Qold−Qold-1)a\left(Q^{\text{old}}-Q^{\text{old-1}}\right).
5: Compute the flux given by eq. eq. 4.
6: Compute the average of the fluxes between QoldQ^{\text{old}} and the adjacent volumes.
7: Update the volume with the fluxes from their 2d adjacent faces subject to a scaling with Δ​tΔ​x\frac{\Delta t}{\Delta x}.
8: Compute the Finite Volume’s non-conservative product.
9: Update the value of QnewQ^{\text{new}}.
10: Compute the patch global maximum eigenvalue.

ExaHyPE offers a generic compute kernel implementation based upon functors, but we also have a templated version handling NN patches in one batch. It consists of a sequence of nested loops (Algorithm 1). We run over all patches. Per patch, we execute different steps, whereas each step consists of loops over cells or faces, respectively, as well as the s=5s=5 unknowns of the PDE.

// Detecting the cartesian product type automatically
using cart_prod_3d = decltype(tl::views::cartesian_product(
std::views::iota(0,1), std::views::iota(0,1),
std::views::iota(0,1)
));
// Allocating in heap
auto* range2d_cell = new cart_prod_3d;
// Defining components
*range2d_cell = tl::views::cartesian_product(
std::views::iota(0,numberOfVolumesPerAxisInPatch),
std::views::iota(0,numberOfVolumesPerAxisInPatch),
std::views::iota(0,hostPatchData.numberOfCells)
);
...
delete[] range2d_cell;
Code 1: Allocation of a Cartesian product (iteration space) on the heap. This way, it can be used by the GPU.

SIMD loop traversals and functors

std::for_each(
std::execution::par_unseq,
range2d_cell->begin(), range2d_cell->end(),
[
=,
QIn=hostPatchData.QIn,
cellCentre=hostPatchData.cellCentre,
cellSize=hostPatchData.cellSize,
t=hostPatchData.t, dt=hostPatchData.dt,
QOut=hostPatchData.QOut
](auto ids){
auto [x,y,patchIndex] = ids;
// function call has to refer to a template, i.e.
// implementation has to be in header. Otherwise,
// compiler does not know that it has to create a
// GPU variant of function.
// step (1) from algorithm blueprint
internal::copySolution_LoopBody<Solver>(
QIn[patchIndex], QInEnumerator,
cellCentre[patchIndex], cellSize[patchIndex],
patchIndex, volumeIndex2d(x,y),
t[patchIndex], dt[patchIndex],
QOut[patchIndex], QOutEnumerator
);
[...]
// steps (2--10) from algorithm blueprint
}
);
Code 2: Standard std::for_each algorithm offloaded to GPU.

The original kernel implementation used virtual-function dynamic dispatch and separately compiled implementations - i.e. in separate .cpp files - to reduce compilation times and improve developer productivity. However, the NVIDIA implementation of ISO C++ does not support virtual-function dispatch in device code and separate compilation. We hence replace virtual-function dispatch in ExaHyPE with template functions defined in header files and static dispatch (Code ). In the case of the maximum in (Algorithm 5), we replace the standard C++ function invocation by a manual comparison.

While the NVIDIA implementation supports separate compilation via non-standard CUDA-like __host__ __device__ annotations, these require the programmer to choose between using a non-standard extension or moving implementation details to header files, which significantly increases compile-times. A manual implementation of maximum is not “genuine C++”.

Multi-dimensional iteration spaces with Cartesian Product

Efficient GPU kernels require large iteration spaces of collapsed loops to harvest all parallelism available in modern accelerators. Hence, a single C++ parallel algorithm that collapses several for loops is built in our Code . It combines multiple std::views::iota structures using C++23’s std::views::cartesian_product. This implementation takes pointers to the host-thread stack. However, according to the HPC SDK documentation [11], only the host heap can be accessed from GPU kernels on the hardware platform used in this study. As a consequence, our code had to allocate the underlying Cartesian product on the heap.

The absence of an implementation of Cartesian product in the C++ standard library used in this study requires us to include a preliminary implementation that is not tuned for accelerators. The heap placement is solely due to technical constraints.

Data transfer

The ISO C++ programming model relies on unified memory to enable the accelerator devices to access all program memory. The NVIDIA implementation is non-conforming on non-coherent hardware platforms like the one used in this study: devices are only allowed to access host heap memory. While ExaHyPE’s data is mainly stored on the heap, pointers to this data are conveniently aggregated in structs like hostPatchData, which are often stored on the host’s stack. Device code is not allowed to access these values to load pointers to heap data. ExaHyPE hence captures these pointers explicitly in the lambdas passed to the parallel algorithms. Lambda captures are copied to the GPU on kernel launch, while the heap data these pointers refer to is migrated on-demand as the GPU threads access it. This data stays on the GPU until the CPU access it again, in which case it is automatically migrated back. There is no portable C++ API for prefetching heap data between the host and the device.

Software stack

Our work uses the NVIDIA HPC SDK 22.3 which leverages the system’s C++ standard library (libstdc++) but uses its own C++ parallel algorithms implementation based on the Thrust library. Many HPC systems do not default to a sufficiently new version of libstdc++, which requires us to manually pick a different one via compiler flags. We reported some genuine nvc++ bugs - e.g.  [12] - which will be solved in future releases.

4 Results

Refer to caption
Figure 1: Time per Finite Volume update for different implementation variants and choices of the number of patches NN handled per kernel invocation.

Fig. 1 shows the benchmark results of our compute kernels for p=17p=17 in a 2d setup on a NVIDIA 2080TI for different sizes of patch sets NN. All timings include all data transfers between host and device. The kernels over NN patches are available as patch-wise implementation where the outermost Cartesian loop runs over the patches, and a batched version, where we run through the individual steps of Algorithm 1 and invoke a loop over all patches, all volumes, and all unknowns per step.

Both kernel variants exploit the symmetries from (4): The face flux enters the left and right adjacent volumes’ solutions. We compute ff (as well as the eigenvalues) per volume and determine the linear combination of these values from (4) redundantly for the left and right adjacent cell. The intermediate ff and λ\lambda values can be held as Array of Structures (AoS), Structure of Arrays (SoA) or Array of Structures of Arrays (AoSoA).

The batched vs. patch-wise flavours result from a permutation of the loops within the Cartesian product (Code ). However, our AMR code employs local and adaptive timestepping and hence requires the maximum eigenvalue per patch as additional output per kernel invocation. We need a per-patch reduction. This reduction is performed by using a sequential std::for_each loop to avoid data races, but it could be parallelized by using another standard algorithm or atomic memory operations.

Our results show that the batched variant is always faster than the patch-wise variant for the C++ implementations. It is advantageous to split an algorithm into its steps and to use a large iteration space per step. As the time per Finite Volume update decreases with the number of patches offloaded, we may assume that the GPU kernel does a good job in keeping the data in-between two steps on the GPU. Data data layout of temporary data within the computation has limited performance impact.

For OpenMP, we see that it can be advantageous to run over the patches one-by-one. We assume that different patches in this case end up on different SMs on the GPU, and run more or less independent of each other. Overall, the OpenMP port is significantly faster—notably for small NN-counts—than the C++ port. NSight studies suggest that this is due to a bespoke OpenMP memory management, where we have explicit control via device pointers when data are offloaded. The C++ version in contrast has the on-demand memory transfers baked into the implementation.

Though our software is ready to allow multiple tasks to offload to the GPU simultaneously, neither C++ nor OpenMP did support this feature. Both ran into segmentation faults.

5 Outlook and conclusion

GPU offloading via C++ execution policies within C++ standard algorithms is, in principle, feasible for non-trivial test cases as our task-based AMR code. The resulting realisation is bare of specific (OpenMP) extensions and clean; though this is a subjective assessment. Performance-wise, the C++ port is slower than its OpenMP counterpart. It is hard to pinpoint a reason for this behaviour. The two-dimensional Euler equations have a very low arithmetic intensity. The performance discrepancies might disappear for more complex calculations such as the ADER-DG ExaHyPE solvers; a phenomenon to be studied in the future.

It is early days to comment on the pros and cons of C++ GPU offloading. Besides obvious teething problems, our work highlights particular challenge for developers: While C++ may be syntactically and semantically correct, its GPU transfer might fail (as data is not allocated “artifically” on the heap, e.g.), or conform by running the code sequentially, violating the intention of the programmer to offload computations to an accelerator. On top of this, the C++ community needs a portable alternative to NVIDIA’s vendor-specific __host__ __device__ annotations for separate compilation. In the ideal case, multi-pass compilation or link-time compilation for GPUs would make these annotations unnecessary.

As alternative to this automatisation, the C++ language might supplement existing execution policies with declarative GPU offloading policies. Using such policies would render otherwise correct loop bodies invalid if they do not fit to the GPU. It is not clear how this is properly reflected within the language, or exclusively should be handled by the compiler throughout translation.

References

  • [1] Volodymyr V. Kindratenko, Jeremy J. Enos, Guochun Shi, Michael T. Showerman, Galen W. Arnold, John E. Stone, James C. Phillips, and Wen-mei Hwu. Gpu clusters for high-performance computing. In 2009 IEEE International Conference on Cluster Computing and Workshops, pages 1–8, 2009.
  • [2] David Olsen. C++17 parallel algorithms on nvidia gpus with pgi c++. In GPU Technology Conference, 2019.
  • [3] Wei-Chen Lin, Tom Deakin, and Simon McIntosh-Smith. Evaluating iso c++ parallel algorithms on heterogeneous hpc systems. In 13th IEEE International Workshop on Performance Modeling, Benchmarking and Simulation of High Performance Computer Systems, 2022.
  • [4] European Exascale Computing Center of Excellence. Research on AI- and Simulation-Based Engineering at Exascale. https://www.hpccoe.eu/2021/06/04/m-aia/, 2021. Last updated: Sep 1, 2022.
  • [5] Jonas Latt, Christophe Coreixas, and Joël Beny. Cross-platform programming model for many-core lattice boltzmann simulations. PLOS ONE, 16(4):e0250306, apr 2021.
  • [6] Anne Reinarz, Dominic E. Charrier, Michael Bader, Luke Bovard, Michael Dumbser, Kenneth Duru, Francesco Fambri, Alice-Agnes Gabriel, Jean-Matthieu Gallard, Sven Köppel, Lukas Krenz, Leonhard Rannabauer, Luciano Rezzolla, Philipp Samfass, Maurizio Tavelli, and Tobias Weinzierl. Exahype: An engine for parallel dynamically adaptive simulations of wave problems. Computer Physics Communications, 254:107251, 2020.
  • [7] Han Zhang, Tobias Weinzierl, Holger Schulz, and Baojiu Li. Spherical accretion of collisional gas in modified gravity I: self-similar solutions and a new cosmological hydrodynamical code. Monthly Notices of the Royal Astronomical Society, 515(2):2464–2482, 07 2022.
  • [8] T. Weinzierl. The peano software—parallel, automaton-based, dynamically adaptive grid traversals. ACM Transactions on Mathematical Software, 45(2):14, 2019.
  • [9] A. Dubey, A. S. Almgren, J. B. Bell, M. Berzins, S. R. Brandt, G. Bryan, P. Colella, D. T. Graves, M. Lijewski, F. Löffler, B. O’Shea, E. Schnetter, B. van Straalen, and K. Weide. A survey of high level frameworks in block-structured adaptive mesh refinement packages. Journal of Parallel and Distributed Computing, 74(12):3217–3227, 2016.
  • [10] Dominic Etienne Charrier, Benjamin Hazelwood, and Tobias Weinzierl. Enclave tasking for dg methods on dynamically adaptive meshes. SIAM Journal on Scientific Computing, 42(3):C69–C96, 2020.
  • [11] NVIDIA HPC Compilers. C++ Parallel Algorithms. https://docs.nvidia.com/hpc-sdk/compilers/c++-parallel-algorithms/index.html, 2022.
  • [12] NVIDIA Bug report. https://nvbugs/3706250, 2022.

Appendix A OpenMP reference implementation

const int numberOfCells = hostPatchData.numberOfCells;
double* rawPointerToCellCentres = hostPatchData.cellCentre[0].data();
double* rawPointerToCellSizes = hostPatchData.cellSize[0].data();
double* t = hostPatchData.t;
double* dt = hostPatchData.dt;
double* maxEigenvalue = hostPatchData.maxEigenvalue;
#pragma omp target enter data map(to:rawPointerToCellCentres[0:hostPatchData.numberOfCells*Dimensions]) device(targetDevice)
#pragma omp target enter data map(to:rawPointerToCellSizes[0:hostPatchData.numberOfCells*Dimensions]) device(targetDevice)
#pragma omp target enter data map(to:mappedPointersToQIn[0:hostPatchData.numberOfCells]) device(targetDevice)
#pragma omp target enter data map(to:mappedPointersToQOut[0:hostPatchData.numberOfCells]) device(targetDevice)
#pragma omp target enter data map(to:t[0:hostPatchData.numberOfCells]) device(targetDevice)
#pragma omp target enter data map(to:dt[0:hostPatchData.numberOfCells]) device(targetDevice)
#pragma omp target enter data map(alloc:maxEigenvalue[0:hostPatchData.numberOfCells]) device(targetDevice)
[...]
#pragma omp target exit data map(delete:rawPointerToCellCentres[0:hostPatchData.numberOfCells*Dimensions]) device(targetDevice)
#pragma omp target exit data map(delete:rawPointerToCellSizes[0:hostPatchData.numberOfCells*Dimensions]) device(targetDevice)
#pragma omp target exit data map(delete:mappedPointersToQIn[0:hostPatchData.numberOfCells]) device(targetDevice)
#pragma omp target exit data map(delete:mappedPointersToQOut[0:hostPatchData.numberOfCells]) device(targetDevice)
#pragma omp target exit data map(delete:t[0:hostPatchData.numberOfCells]) device(targetDevice)
#pragma omp target exit data map(delete:dt[0:hostPatchData.numberOfCells]) device(targetDevice)
#pragma omp target exit data map(from:maxEigenvalue[0:hostPatchData.numberOfCells]) device(targetDevice)
Code 3: Allocation and deallocation of variables on the device for OpenMP.

Our OpenMP vanilla version hosts the data of NN (number of cells) patches within a struct on the host. Data here comprises the actual Finite Volume data as well as meta data such as cell (patch) sizes and positions.

The OpenMP variant maps these data explicitly onto the GPU prior to the kernel invocation (Code ). An analogous loop over the individual patches transfers the Finite Volume data per patch (now shown).

#pragma omp target teams distribute parallel for simd collapse(3) device(targetDevice)
for (int patchIndex=0; patchIndex<numberOfCells; patchIndex++)
for (int y = 0; y < numberOfVolumesPerAxisInPatch; y++)
for (int x = 0; x < numberOfVolumesPerAxisInPatch; x++) {
GPUCellData patchData;
patchData.numberOfCells = numberOfCells;
patchData.cellCentre = (tarch::la::Vector<Dimensions,double>*)(rawPointerToCellCentres);
patchData.cellSize = (tarch::la::Vector<Dimensions,double>*)(rawPointerToCellSizes);
patchData.QIn = mappedPointersToQIn;
patchData.QOut = mappedPointersToQOut;
patchData.t = t;
patchData.dt = dt;
patchData.maxEigenvalue = maxEigenvalue;
[...]
copySolution_LoopBody<Solver>(
patchData.QIn[patchIndex],
QInEnumerator,
patchData.cellCentre[patchIndex],
patchData.cellSize[patchIndex],
patchIndex,
volumeIndex2d(x,y),
patchData.t[patchIndex],
patchData.dt[patchIndex],
patchData.QOut[patchIndex],
QOutEnumerator
);
[...]
}
Code 4: Main loop body for the OpenMP code. Equivalent to Code .

Our OpenMP reference implementation realises the nested Cartesian iteration space via pure nested for loops (Code ). They are collapsed via a pragma. In this sense, the realisation is closer to C than its C++ GPU offloading counterpart.

Within the loop body, the code reconstructs the host struct manually using device pointers, i.e. it creates a struct counterpart on the GPU and befills its translated GPU pointers manually. Alternatively, a realisation could have supplemented the struct hosting the NN patch data with an OpenMP mapping.

The function copySolutionAndAddSourceTerm_LoopBody and other compute step realisations from Algorithm 1 are templates and therefore found in headers. Consequently, the compiler can directly inline them and deduce from the context that they have to be translated for the GPU.

Appendix B Test environment

We provide a docker image to reproduce the results obtained in this paper. The custom image is based on the latest HPC SDK compiler, and it includes some Python libraries and other necessary software to run ExaHyPE and its underlying Peano AMR framework. If the software is going to be run on a GPU, use the following command to pull the image and start a bash terminal

docker run --gpus all -it ghcr.io/uzmargomez/peano:hpcsdk22.9 bash

Without access to a GPU, the --gpus all flag has to be omitted. In the container’s terminal, we next run the commands

cd /
git clone https://gitlab.lrz.de/hpcsoftware/Peano.git && cd Peano && git checkout gpus && git checkout c57ef3b5
export PYTHONPATH=/Peano/python
libtoolize && aclocal && autoconf && autoheader && cp src/config.h.in . && automake --add-missing
./configure --enable-exahype --enable-blockstructured --enable-loadbalancing --with-gpu=cpp --with-multithreading=cpp CXX=$NVCPP CC=$NVCPP CXXFLAGS="-O4 -std=c++20 -stdpar=gpu -gpu=cc75" LDFLAGS="-stdpar=gpu -gpu=cc75"
make -j4
cd benchmarks/exahype2/euler/fv-kernel-benchmarks
python3 euler-fv-rusanov-kernel-benchmarks.py -a 1e-4
make -j8
./fv-kernel-benchmarks

If a GPU is not available, the HPC SDK compiler can use a multicore CPU to parallelize standard algorithms. In this case, the image’s application benchmarks solely the CPU kernel performance. In order to do this, we change the flags -stdpar=gpu to -stdpar=multicore.